"""All definitions for lidia configuration, reused to generate the schema for config files.

To support setting it remotely when using MARSH source, specific numeric and boolean values
should have names that are unique and no longer than 16 characters.
"""
from enum import Enum
import os
from os import path
from typing import Dict, Optional
from pydantic import Field

from .mytypes import NestingModel


class RpctaskConfig(NestingModel):
    """Configuration for `rpctask`"""
    ok_tolerance: float = Field(default=0.03, alias='correct_tolerance')
    """Acceptable margin for green color.

DEPRECATED: Renamed to `ok_tolerance`, `correct_tolerance` won't be accepted after next major release"""
    warn_tolerance: float = Field(default=0.05, alias='warning_tolerance')
    """Acceptable margin for yellow color.

DEPRECATED: Renamed to `warn_tolerance`, `warning_tolerance` won't be accepted after next major release"""


class PfdConfig(NestingModel):
    """configuration for `pfd`"""
    show_alt_target: bool = True
    """Display the altitude target after receiving `trgt.instr.alt`"""
    ias_never_exceed: float = 167.0
    """Never exceed speed (Vne) shown on IAS tape"""
    show_ias_target: bool = True
    """Display the IAS target after receiving `trgt.instr.ias`"""
    show_vsi_target: bool = True
    """Display the VSI target after receiving `trgt.v_ned`"""
    show_adi_target: bool = True
    """Display the attitude target after receiving `trgt.att`"""
    adi_target_roll: bool = True
    """Rotate attitude target indicator to show desired roll"""
    adi_target_yaw: bool = False
    """Move attitude target indicator to show desired yaw"""
    show_flightpath: bool = True
    """Show flightpath vector (FPV) indicator on ADI"""
    show_retrograde: bool = False
    """Show reverse flight path vector on ADI"""
    move_roll_ticks: bool = False
    """Move roll angle scale with horizon, keeping sideslip triangle in place"""
    sideslip_max: float = 15.0
    """Maximal displayed sideslip angle, in degrees"""
    rpm_e_good_low: float = 0.97
    """Engine RPM safe region lower limit"""
    rpm_e_good_high: float = 1.03
    """Engine RPM safe region upper limit"""
    rpm_e_warn_low: float = 0.97
    """Engine RPM warning region lower limit"""
    rpm_e_warn_high: float = 1.03
    """Engine RPM warning region upper limit"""
    rpm_e_nominal: float = 6000.0
    """Engine RPM nominal value, in revolutions per minute"""
    rpm_r_good_low: float = 0.97
    """Rotor RPM safe region lower limit"""
    rpm_r_good_high: float = 1.03
    """Rotor RPM safe region upper limit"""
    rpm_r_warn_low: float = 0.90
    """Rotor RPM warning region lower limit"""
    rpm_r_warn_high: float = 1.10
    """Rotor RPM warning region upper limit"""
    rpm_r_nominal: float = 380.0
    """Rotor RPM nominal value, in revolutions per minute"""
    traffic_range: float = 18520.0
    """Range of displayed traffic information, in meters"""


class ApproachConfig(NestingModel):
    """Ship approach configuration"""
    nominal_alt: float = 3
    """Altitude at which the scale is 1, in meters"""
    camera_height: float = 10
    """Position of camera above aircraft origin, in meters

    Larger values of this make the scale change less drastically at low altitude"""


class InstrumentsConfig(NestingModel):
    """Configuration for instruments visualisation, units etc."""
    speed_multiplier: float = 3600.0 / 1852.0
    """Scaling factor to change state velocity in meters per second to displayed IAS and GS, default for knots"""
    alt_multiplier: float = Field(
        default=1 / 0.3048, alias='altitude_multiplier')
    """Scaling factor to change state altitude in meters to displayed altitude, default for feet

DEPRECATED: Renamed to `alt_multiplier`, `altitude_multiplier` won't be accepted after next major release"""
    ralt_activation: float = Field(
        default=2500.0 * 0.3048, alias='radio_altimeter_activation')
    """Activation height of radio altimeter above which it is not modeled, default 2500ft

DEPRECATED: Will be renamed to `ralt_activation` in next major release"""


class CASCategory(Enum):
    """Category (severity, type) of a message shown in CAS"""
    WARNING = 1
    """Red, blinking until acknowledged"""
    CAUTION = 2
    """Amber, blinking until acknowledged"""
    ADVISORY = 3
    """Green, blinking for a fixed time"""
    STATUS = 4
    """White messages"""


class CASEvent(NestingModel):
    """Event that can be displayed in CAS"""
    category: CASCategory
    """Category to put the message into"""
    text: str
    """Text to be shown"""


class CASConfig(NestingModel):
    """Configuration for Crew Alerting System, including events"""
    events: Dict[int, CASEvent] = {}
    """Dictionary of events by their integer id"""


class Config(NestingModel):
    """Root of configuration structure

    Every config field in this and children should be provided with defaults,
    that can be overriden by config files"""

    json_schema_url: Optional[str] = Field(alias='$schema', default=None)
    """Allow the `$schema` property for specifying JSON Schema URL"""
    rpctask = RpctaskConfig()
    pfd = PfdConfig()
    approach = ApproachConfig()
    instruments = InstrumentsConfig()
    cas = CASConfig()
    start_time: Optional[float] = None
    """Epoch time in seconds of starting the program (see `time.time()`)"""


def schema_location():
    root_path = path.abspath(path.dirname(__file__))
    data_path = path.join(root_path, 'data')
    os.makedirs(data_path, exist_ok=True)
    return path.join(data_path, 'lidia-config.json')


def write_schema():
    with open(schema_location(), 'w') as out:
        out.write(Config.schema_json(by_alias=True))


if __name__ == '__main__':
    # check that config works with old and new names
    print(Config().updated({"rpctask": {"correct_tolerance": 0.01}}).rpctask)
    print(Config().updated({"rpctask": {"ok_tolerance": 0.01}}).rpctask)
    # still rejects wrong fields
    try:
        print(Config().updated({"rpctask": {"foo": 0.01}}).rpctask)
    except Exception as e:
        print(e)



from argparse import Namespace
import queue
import eventlet
from msgpack import packb
from multiprocessing import Queue
from os import path
import socket
import socketio
from typing import Dict, Optional


from .config import Config


def background_loop(q: Queue, sio: socketio.Server, args: Namespace):
    udp_host, udp_port = args.passthrough_host, args.passthrough_port
    udp_sock = None
    if udp_host is not None and udp_port is not None:
        udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    while True:
        try:
            while not q.empty():
                event, data = q.get()
                sio.emit(event, data)
                if args.verbosity >= 2:
                    print(f"socketio.emit('{event}', {data})")
                if event == 'smol' and udp_sock is not None:
                    sent_len = udp_sock.sendto(
                        packb(data), (udp_host, udp_port))
                    if args.verbosity >= 2:
                        print(
                            f"sent {sent_len} bytes to udp:{udp_host}:{udp_port}")
        except queue.Empty:
            pass
        eventlet.sleep(0.02)


def run_server(q: Queue, args: Namespace, config: Config):
    # specifying just local path breaks when run as a module
    root_path = path.abspath(path.dirname(__file__))
    pages_path = path.join(root_path, 'pages')
    static_files = {
        # @deprecated: change to info.html
        '/': {'content_type': 'text/html', 'filename': path.join(pages_path, 'controls.html')},
        '/info': {'content_type': 'text/html', 'filename': path.join(pages_path, 'info.html')},
        '/controls': {'content_type': 'text/html', 'filename': path.join(pages_path, 'controls.html')},
        '/pfd': {'content_type': 'text/html', 'filename': path.join(pages_path, 'pfd.html')},
        '/approach': {'content_type': 'text/html', 'filename': path.join(pages_path, 'approach.html')},
        '/static': path.join(root_path, 'pages/static'),
    }

    sio = socketio.Server()

    @sio.on('config_request')
    def config_request(_sid, _environ):
        q.put(('config', config.dict()))

    app = RedirectingWSGIApp(sio, static_files=static_files, redirects={
        '/rpctask': '/controls'
    })
    eventlet.spawn(background_loop, q, sio, args)
    eventlet.wsgi.server(eventlet.listen((args.http_host, args.http_port)),
                         app, log_output=args.verbosity >= 3)


class RedirectingWSGIApp(socketio.WSGIApp):
    
    def __init__(self, socketio_app, static_files=None, redirects: Optional[Dict[str, str]]=dict()):
        super(RedirectingWSGIApp, self).__init__(socketio_app, static_files=static_files)
        self.redirects = redirects

    def __call__(self, environ, start_response):
        path = environ['PATH_INFO']
        if path in self.redirects:
            start_response('301 Moved Permanently', [
                ('Location', self.redirects[path]),
                ('Cache-Control', 'max-age=2592000'),  # 30 days, from example on MDN
                ('Content-Type', 'text/plain'),
            ])
            return [f'Path {path} is deprecated, redirecting to {self.redirects[path]}'.encode()]

        return super(RedirectingWSGIApp, self).__call__(environ, start_response)



"""connect to MARSH Sim system, see https://marsh-sim.github.io/"""
from argparse import _SubParsersAction, ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from collections import OrderedDict
from math import isnan, nan, sqrt
import msgpack
from multiprocessing import Queue
from pydantic.fields import ModelField
from pymavlink import mavutil
from time import time
from typing import Tuple

from ..aircraft import *
from ..config import Config
from .. import mavlink_all as mavlink
from ..mytypes import NestingModel
from .mytypes import RunFn


def setup(subparsers: _SubParsersAction) -> Tuple[str, RunFn]:
    NAME = 'marsh'
    parser: ArgumentParser = subparsers.add_parser(
        NAME,
        help=__doc__,
        formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-m', '--manager',
                        help='MARSH Manager IP addr', default='127.0.0.1')

    return (NAME, run)


class ParamConfig:
    def __init__(self):
        self.info = OrderedDict()
        self.ignore_fields = {
            'json_schema_url',
            'cas',
        }
        for name, field in Config.__fields__.items():
            if name in self.ignore_fields:
                continue

            self._add_field_info([name], field)

    def _add_field_info(self, path: List[str], field: ModelField):
        if issubclass(field.type_, NestingModel):
            model: NestingModel = field.type_
            for name, inner_field in model.__fields__.items():
                if name in self.ignore_fields:
                    continue
                self._add_field_info(path + [name], inner_field)
        elif field.type_ in {int, float, bool}:
            assert path[-1] not in self.info
            self.info[path[-1]] = (path, field.type_)

    def get(self, config: Config, param_name: str):
        path, _ = self.info[param_name]
        field = config
        for part in path:
            field = getattr(field, part)
        return field

    def set(self, config: Config, param_name: str, value: float | int | bool):
        path, type_ = self.info[param_name]
        if type_ is bool and value is not bool:
            assert value == 0 or value == 1
        elif type_ is int:
            assert round(value) == value

        obj = config
        for part in path[:-1]:
            obj = getattr(obj, part)
        setattr(obj, path[-1], type_(value))

    def __contains__(self, item):
        return item in self.info

    def __len__(self):
        return len(self.info)

    def send(self, config: Config, mav: mavlink.MAVLink, index: int, name=''):
        """
        convenience function to send PARAM_VALUE
        pass index -1 to use name instead

        silently returns on invalid index or name
        """
        param_id = bytearray(16)

        if index >= 0:
            if index >= len(self):
                return

            # HACK: is there a nicer way to get items from OrderedDict by order?
            name = list(self.info.keys())[index]
        else:
            if name not in self:
                return

            index = list(self.info.keys()).index(name)
        name_bytes = name.encode('utf8')
        param_id[:len(name_bytes)] = name_bytes

        mav.param_value_send(param_id, self.get(config, name), mavlink.MAV_PARAM_TYPE_REAL32,
                             len(self), index)


def run(q: Queue, args: Namespace, config: Config):

    connection_string = f'udpout:{args.manager}:24400'
    mav = mavlink.MAVLink(mavutil.mavlink_connection(connection_string))
    mav.srcSystem = 1  # default system
    mav.srcComponent = mavlink.MAV_COMP_ID_USER1 + (mavlink.MARSH_TYPE_INSTRUMENTS - mavlink.MARSH_TYPE_MANAGER)
    if args.verbosity >= 0:
        print(f'Connecting to MARSH Manager on {connection_string}')

    last_state = AircraftState()
    params = ParamConfig()

    # for messages requiring time_boot*
    start_time = time()

    # controlling when messages should be sent
    heartbeat_next = 0.0
    heartbeat_interval = 1.0

    # monitoring connection to manager with heartbeat
    timeout_interval = 5.0
    manager_timeout = 0.0
    manager_connected = False
    manager_component = mavlink.MAV_COMP_ID_USER1  # this is only the default

    # the loop goes as fast as it can, relying on the variables above for timing
    while True:
        if time() >= heartbeat_next:
            mav.heartbeat_send(
                mavlink.MARSH_TYPE_INSTRUMENTS,
                mavlink.MAV_AUTOPILOT_INVALID,
                mavlink.MAV_MODE_FLAG_TEST_ENABLED,
                0,
                mavlink.MAV_STATE_ACTIVE
            )
            heartbeat_next = time() + heartbeat_interval

        state_changed = False

        # handle incoming messages
        try:
            while (message := mav.file.recv_msg()) is not None:
                message: mavlink.MAVLink_message
                if args.verbosity >= 2:
                    print('Received', message.get_type())
                if message.get_type() == 'HEARTBEAT':
                    heartbeat: mavlink.MAVLink_heartbeat_message = message
                    if heartbeat.type == mavlink.MARSH_TYPE_MANAGER:
                        manager_component = message.get_srcComponent()
                        if not manager_connected and args.verbosity >= 0:
                            print('Connected to simulation manager')
                            for msgid in [
                                mavlink.MAVLINK_MSG_ID_V2_EXTENSION,
                                mavlink.MAVLINK_MSG_ID_SET_POSITION_TARGET_LOCAL_NED,
                            ]:
                                mav.command_long_send(mav.srcSystem, manager_component,
                                                      mavlink.MAV_CMD_SET_MESSAGE_INTERVAL,
                                                      0,  # not a confirmation
                                                      msgid,  # requested msg
                                                      0,  # default interval
                                                      nan, nan, nan, nan,  # unused
                                                      1)  # request is for sender (lidia)
                        manager_connected = True
                        manager_timeout = time() + timeout_interval
                elif message.get_type() == 'MANUAL_CONTROL':
                    # This line helps with type hints
                    mc: mavlink.MAVLink_manual_control_message = message
                    state_changed = True

                    last_state.ctrl = Controls()
                    # Invalid axes are sent as INT16_MAX
                    INVALID = 0x7FFF
                    if mc.x != INVALID:
                        last_state.ctrl.stick_pull = mc.x / -1000
                    if mc.y != INVALID:
                        last_state.ctrl.stick_right = mc.y / 1000
                    if mc.z != INVALID:
                        last_state.ctrl.throttle = mc.z / 1000
                        last_state.ctrl.collective_up = mc.z / 1000
                    if mc.z != INVALID:
                        last_state.ctrl.pedals_right = mc.r / 1000

                    # TODO: assign buttons

                elif message.get_type() == 'MANUAL_SETPOINT':
                    # This line helps with type hints
                    ms: mavlink.MAVLink_manual_setpoint_message = message  # pyright:ignore[reportAssignmentType]
                    state_changed = True

                    if ms.mode_switch == 0:  # MARSH_MANUAL_SETPOINT_MODE_TARGET
                        if last_state.trgt is None:
                            last_state.trgt = AircraftData()
                        last_state.trgt.ctrl = Controls()
                        last_state.trgt.ctrl.stick_right = ms.roll
                        last_state.trgt.ctrl.stick_pull = ms.pitch
                        last_state.trgt.ctrl.pedals_right = ms.yaw
                        last_state.trgt.ctrl.throttle = ms.thrust
                        last_state.trgt.ctrl.collective_up = ms.thrust
                    elif ms.mode_switch == 1:  # MARSH_MANUAL_SETPOINT_MODE_TRIM
                        if last_state.trim is None:
                            last_state.trim = AircraftData()
                        last_state.trim.ctrl = Controls()
                        last_state.trim.ctrl.stick_right = ms.roll
                        last_state.trim.ctrl.stick_pull = ms.pitch
                        last_state.trim.ctrl.pedals_right = ms.yaw
                        last_state.trim.ctrl.throttle = ms.thrust
                        last_state.trim.ctrl.collective_up = ms.thrust
                    else:
                        if args.verbosity >= 1:
                            print(f'Unhandled MANUAL_SETPOINT.mode_switch {ms.mode_switch}, check MARSH_MANUAL_SETPOINT_MODE definitions')

                elif message.get_type() == 'SIM_STATE':
                    ss: mavlink.MAVLink_sim_state_message = message
                    state_changed = True

                    last_state.att = Attitude()
                    last_state.att.roll = ss.roll
                    last_state.att.pitch = ss.pitch
                    last_state.att.yaw = ss.yaw

                    last_state.a_body = XYZ()
                    last_state.a_body.x = ss.xacc
                    last_state.a_body.y = ss.yacc
                    last_state.a_body.z = ss.zacc

                    last_state.v_ned = NED()
                    last_state.v_ned.north = ss.vn
                    last_state.v_ned.east = ss.ve
                    last_state.v_ned.down = ss.vd

                    last_state.ned = NED()
                    # TODO: Add argument for reference location, calculate north and east from that
                    last_state.ned.down = -ss.alt

                elif message.get_type() == 'LOCAL_POSITION_NED':
                    lpn: mavlink.MAVLink_local_position_ned_message = message
                    state_changed = True

                    last_state.ned = NED()
                    last_state.ned.north = lpn.x
                    last_state.ned.east = lpn.y
                    last_state.ned.down = lpn.z

                    last_state.v_ned = NED()
                    last_state.v_ned.north = lpn.vx
                    last_state.v_ned.east = lpn.vy
                    last_state.v_ned.down = lpn.vz

                elif message.get_type() == 'ATTITUDE':
                    att: mavlink.MAVLink_attitude_message = message
                    state_changed = True

                    last_state.att = Attitude()
                    last_state.att.roll = att.roll
                    last_state.att.pitch = att.pitch
                    last_state.att.yaw = att.yaw

                elif message.get_type() == 'HIGHRES_IMU':
                    imu: mavlink.MAVLink_highres_imu_message = message
                    state_changed = True

                    last_state.a_body = XYZ()
                    last_state.a_body.x = imu.xacc
                    last_state.a_body.y = imu.yacc
                    last_state.a_body.z = imu.zacc

                elif message.get_type() == 'RAW_RPM':
                    rr: mavlink.MAVLink_raw_rpm_message = message
                    state_changed = True

                    if last_state.hrpm is None:
                        last_state.hrpm = HelicopterRPM()

                    if rr.index == 0:
                        last_state.hrpm.rotor = rr.frequency / config.pfd.rpm_r_nominal
                    else:
                        last_state.hrpm.engine = rr.frequency / config.pfd.rpm_e_nominal

                elif message.get_type() == 'SET_POSITION_TARGET_LOCAL_NED':
                    local: mavlink.MAVLink_set_position_target_local_ned_message = message
                    if last_state.trgt is None:
                        last_state.trgt = AircraftData()
                    if last_state.trgt.instr is None:
                        last_state.trgt.instr = Instruments()

                    # ignore all, don't use force instead of acceleration
                    type_mask = 0b0000_1101_1111_1111

                    alt = -local.z
                    if local.type_mask & mavlink.POSITION_TARGET_TYPEMASK_Z_IGNORE:
                        alt = nan
                    if not isnan(alt):
                        last_state.trgt.instr.alt = alt * config.instruments.alt_multiplier
                        type_mask ^= mavlink.POSITION_TARGET_TYPEMASK_Z_IGNORE  # flip the bit to not ignored
                        state_changed = True

                    speed = local.vx
                    if not isnan(local.vy):
                        speed = sqrt(local.vx ** 2 + local.vy ** 2)
                    if local.type_mask & mavlink.POSITION_TARGET_TYPEMASK_VX_IGNORE:
                        speed = nan
                    if not isnan(speed):
                        last_state.trgt.instr.ias = speed * config.instruments.speed_multiplier
                        # flip the bit to not ignored
                        type_mask ^= mavlink.POSITION_TARGET_TYPEMASK_VX_IGNORE
                        state_changed = True

                    climb = -local.vz
                    if local.type_mask & mavlink.POSITION_TARGET_TYPEMASK_VZ_IGNORE:
                        climb = nan
                    if not isnan(climb):
                        if last_state.trgt.v_ned is None:
                            last_state.trgt.v_ned = NED()
                        last_state.trgt.v_ned.down = climb
                        # flip the bit to not ignored
                        type_mask ^= mavlink.POSITION_TARGET_TYPEMASK_VZ_IGNORE
                        state_changed = True

                    yaw = local.yaw
                    if local.type_mask & mavlink.POSITION_TARGET_TYPEMASK_YAW_IGNORE:
                        yaw = nan
                    if not isnan(yaw):
                        if last_state.trgt.att is None:
                            last_state.trgt.att = Attitude()
                        last_state.trgt.att.yaw = yaw
                        # flip the bit to not ignored
                        type_mask ^= mavlink.POSITION_TARGET_TYPEMASK_YAW_IGNORE
                        state_changed = True

                    # Respond with the values that were set by this message
                    mav.position_target_local_ned_send(
                        round((time() - start_time) *
                              1000), mavlink.MAV_FRAME_LOCAL_NED, type_mask,
                        nan, nan, alt,
                        nan, nan, climb,
                        nan, nan, nan,
                        yaw, nan
                    )

                elif message.get_type() in ['PARAM_REQUEST_READ', 'PARAM_REQUEST_LIST', 'PARAM_SET']:
                    # check that this is relevant to us
                    if message.target_system == mav.srcSystem and message.target_component == mav.srcComponent:
                        if message.get_type() == 'PARAM_REQUEST_READ':
                            m: mavlink.MAVLink_param_request_read_message = message
                            params.send(config, mav, m.param_index, m.param_id)
                        elif message.get_type() == 'PARAM_REQUEST_LIST':
                            for i in range(len(params)):
                                params.send(config, mav, i)
                        elif message.get_type() == 'PARAM_SET':
                            m: mavlink.MAVLink_param_set_message = message
                            # check that parameter is defined and sent as float
                            if m.param_id in params and m.param_type == mavlink.MAV_PARAM_TYPE_REAL32:
                                params.set(config, m.param_id, m.param_value)
                                q.put(('config', config.dict(by_alias=True)))
                            params.send(config, mav, -1, m.param_id)

                elif message.get_type() == 'V2_EXTENSION':
                    # smol payload passthrough
                    ext: mavlink.MAVLink_v2_extension_message = message
                    if ext.message_type == 44400:  # may change to 24400 if registered
                        try:
                            # some implementations don't handle payload truncation well,
                            # this way only unpacks the first object and ignores any trailing data
                            unpacker = msgpack.Unpacker()
                            unpacker.feed(bytes(ext.payload))
                            decoded = next(unpacker)
                            state = AircraftState.from_smol(decoded)

                            state_changed = last_state.update(state, True)
                        except Exception as e:
                            print('Error in V2_EXTENSION payload:', e)

        except ConnectionResetError:
            # thrown on Windows when there is no peer listening
            pass

        if manager_connected and time() > manager_timeout:
            manager_connected = False
            if args.verbosity >= 0:
                print('Lost connection to simulation manager')

        if state_changed:
            state = last_state
            state.model_instruments(config)
            if state.trgt is not None:
                state.trgt.model_instruments(config)
            if state.trim is not None:
                state.trim.model_instruments(config)
            q.put(('smol', state.smol()))


if __name__ == '__main__':
    # try accessing config with strings
    config = Config()
    print(getattr(getattr(config, 'rpctask'), 'ok_tolerance'))
    setattr(getattr(config, 'rpctask'), 'ok_tolerance', 0.02)
    print(getattr(getattr(config, 'rpctask'), 'ok_tolerance'))

    # check accessing through ParamConfig
    param = ParamConfig()
    print(param.get(config, 'ok_tolerance'))
    param.set(config, 'ok_tolerance', 0.01)
    print(config.rpctask.ok_tolerance)



import argparse
from enum import auto, IntEnum
import inspect
from io import StringIO, TextIOBase
import os
import platform
from pydantic import BaseModel
import re
import sys
from typing import Any, Dict, List, Tuple, Type

# setup cross-platform getch() and clear()
if platform.system() == "Windows":
    import msvcrt

    def getch():
        return msvcrt.getwch()

    def clear():
        os.system('cls')
else:
    import tty
    import termios

    def getch():
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch

    def clear():
        os.system('clear')

this_path = os.path.dirname(os.path.realpath(__file__))
lidia_path = os.path.join(this_path, '..', '..')
sys.path.append(lidia_path)
from lidia.aircraft import AircraftData, VectorModel  # noqa prevent moving to top of file


class FieldInfo(BaseModel):
    """Details of a `AircraftData` field"""

    name: str
    """Field name"""
    docstring: str
    """Field docstring"""
    field_type: Type
    """Contained type"""

    @staticmethod
    def inspect_AircraftData() -> List['FieldInfo']:
        infos = []
        pattern = re.compile(
            r'^[ \t]+([a-z_]+): Optional\[(?:[A-Za-z]+)\] = None\n[ \t]+"""([^"]+)"""$', flags=re.MULTILINE)
        for name, doc in re.findall(pattern, inspect.getsource(AircraftData)):
            field = AircraftData.__fields__[name]
            infos.append(FieldInfo(
                name=name,
                docstring=doc,
                field_type=field.annotation.__args__[
                    0],  # annotation type Optional
            ))

        return infos


class Generator(IntEnum):
    MATLAB = auto()
    MBDYN = auto()
    MBDYN_TYPE = auto()


class Model:
    def __init__(self, output: str, verbose=False):
        self.output = output
        self.editing_output = False
        self.previous_output = ''
        self.verbose = verbose
        self.last_key = 0

        self.parts = ['main', 'trgt', 'trim']
        self.selected_part = self.parts[0]

        # only serialization of vectors is supported
        self.toggles = [info for info in FieldInfo.inspect_AircraftData()
                        if (issubclass(info.field_type, VectorModel) or info.field_type is int)]
        self.enabled = {part: ([False] * len(self.toggles))
                        for part in self.parts}

        self.generators = [
            (Generator.MATLAB, 'MATLAB', '.m'),
            (Generator.MBDYN, 'MBDyn', '{.set, .drv, .elm}'),
        ]
        self.enabled_gens = [False] * len(self.generators)

        self.extra_choices = 3
        self.selected_choice = 0

        self.status = ''

    @property
    def choices_num(self) -> int:
        return len(self.toggles) + len(self.parts) + len(self.generators) + self.extra_choices

    @property
    def selected_part_index(self) -> int:
        return next(i for i, p in enumerate(self.parts) if p == self.selected_part)

    @property
    def enabled_num(self) -> int:
        return sum(len([en for en in self.enabled[p] if en]) for p in self.parts)


class Command(IntEnum):
    QUIT = auto()
    MAKE = auto()


class Message(IntEnum):
    KEY = auto()
    STATUS = auto()


def view(model: Model) -> None:
    clear()
    if model.editing_output:
        print('Edit output filename')
        print('Confirm editing with Enter, cancel with Esc\n')
    else:
        print('Choose fields to be serialized in message')
        print('Use arrows or j, k to move, select with Enter or Space\n')

    for i, (info, enabled) in enumerate(zip(model.toggles, model.enabled[model.selected_part])):
        print('{}({}) {:<16}{} ({})'.format(
            '->' if i == model.selected_choice else '  ',
            '*' if enabled else ' ',
            info.name if model.selected_part == model.parts[0] else '{}.{}'.format(
                model.selected_part, info.name),
            info.docstring,
            info.field_type.__name__))
    print()
    for i, part in enumerate(model.parts):
        print('{}({}) {} edit {} message'.format(
            '->' if i == model.selected_choice - len(model.toggles) else '  ',
            '*' if part == model.selected_part else ' ',
            '[Tab]' if i == (model.selected_part_index +
                             1) % len(model.parts) else '     ',
            part))
    print()
    for i, (_, name, extension) in enumerate(model.generators):
        print('{}({}) {}: {}{}'.format(
            '->' if i == model.selected_choice -
            len(model.toggles) - len(model.parts) else '  ',
            '*' if model.enabled_gens[i] else ' ',
            name,
            model.output,
            extension))

    print('\n{}[o]utput file: {}'.format(
        '->' if model.selected_choice == model.choices_num - 3 else '  ',
        model.output + ('█' if model.editing_output else '')))
    gen_count = model.enabled_gens.count(True)
    field_count = sum([model.enabled[p].count(True) for p in model.parts])
    print('{}[g]enerate with {} generator{} for {} field{}'.format(
        '->' if model.selected_choice == model.choices_num - 2 else '  ',
        gen_count, '' if gen_count == 1 else 's',
        field_count, '' if field_count == 1 else 's',
    ))
    print('{}[q]uit'.format(
        '->' if model.selected_choice == model.choices_num - 1 else '  '))

    print('\n{}'.format(model.status))
    if model.verbose:
        print('last key pressed {}'.format(model.last_key))


def update(model: Model, message: Message, data: Any) -> Tuple[Model, List[Command]]:
    commands = []
    if message == Message.KEY:
        key: int = data
        model.last_key = key

        if model.editing_output:
            model = update_output(model, key)
        else:
            model, cs = update_menu(model, key)
            commands.extend(cs)

    elif message == Message.STATUS:
        status: str = data
        model.status = status

    return model, commands


def update_output(model: Model, key: int) -> Model:
    if key >= 32 and key <= 126:  # printable ASCII characters
        model.output += chr(key)
    elif key in [8, 127]:  # Backspace (Win, Linux)
        model.output = model.output[:-1]
    elif key == 13:  # Enter
        if model.output == '':
            model.status = 'Empty output filename!'
        else:
            model.editing_output = False
            model.status = ''
    elif key == 27:  # Esc
        model.output = model.previous_output
        model.editing_output = False
        model.status = 'Edit cancelled'
    return model


def update_menu(model: Model, key: int) -> Tuple[Model, List[Command]]:
    commands = []
    if key in [ord('o'), ord('O')]:
        model.selected_choice = model.choices_num - 3
        model.previous_output = model.output
        model.editing_output = True
    elif key in [ord('g'), ord('G')]:
        model.selected_choice = model.choices_num - 2
        commands.append(Command.MAKE)
    elif key in [ord('q'), ord('Q'), 3]:  # Ctrl+C
        commands.append(Command.QUIT)
    elif key in [ord('j'), ord('J'), 66, 80]:  # Linux down, Win down
        model.selected_choice = (
            model.selected_choice + 1) % model.choices_num
    elif key in [ord('k'), ord('K'), 65, 72]:  # Linux up, Win up
        model.selected_choice = (
            model.selected_choice - 1) % model.choices_num
    elif key == 9:  # Tab
        model.selected_part = model.parts[(
            model.selected_part_index + 1) % len(model.parts)]

    elif key in [13, 32]:  # Enter, Space
        i = model.selected_choice
        if 0 <= i < len(model.toggles):
            model.enabled[model.selected_part][i] = not model.enabled[model.selected_part][i]
        i -= len(model.toggles)

        if 0 <= i < len(model.parts):
            model.selected_part = model.parts[i]
        i -= len(model.parts)

        if 0 <= i < len(model.generators):
            model.enabled_gens[i] = not model.enabled_gens[i]
        i -= len(model.generators)

        if model.selected_choice == model.choices_num - 3:
            model.previous_output = model.output
            model.editing_output = True
        elif model.selected_choice == model.choices_num - 2:
            commands.append(Command.MAKE)
        elif model.selected_choice == model.choices_num - 1:
            commands.append(Command.QUIT)

    return model, commands


def main():
    parser = argparse.ArgumentParser(
        description='generate MATLAB function to pack data for lidia',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose',
                        help='increase amount of shown information')
    parser.add_argument('-o', '--output', default='pack_lidia',
                        help='output file path')
    args = parser.parse_args()

    model = Model(os.path.normpath(args.output), args.verbose)
    messages = []
    while True:
        commands = []
        for msg, data in messages:
            mdl, c = update(model, msg, data)
            model = mdl
            commands.extend(c)
            messages.clear()

        for cmd in commands:
            if cmd == Command.QUIT:
                sys.exit(0)
            if cmd == Command.MAKE:
                selected = {p: [info.name for info, en in zip(
                    model.toggles, model.enabled[p]) if en] for p in model.parts}
                if len(os.path.dirname(model.output)) > 0:
                    os.makedirs(os.path.dirname(model.output), exist_ok=True)
                filename = os.path.basename(model.output)

                status = []
                infos = {
                    info.name: info for info in FieldInfo.inspect_AircraftData()}
                for enabled, (gen, name, extension) in zip(model.enabled_gens, model.generators):
                    if not enabled:
                        continue

                    out_path = model.output + extension
                    if gen == Generator.MATLAB:
                        forbidden_characters = re.findall(
                            r'[^A-Za-z0-9_]', model.output)
                        if len(forbidden_characters) > 0:
                            status.append(
                                f'Error: characters {set(forbidden_characters)} not allowed in MATLAB function name')
                        else:
                            with open(out_path, 'w') as out:
                                codegen_matlab(out, filename, infos,
                                               selected['main'], selected['trgt'], selected['trim'])
                            status.append(
                                f'Saved {name} generator to {out_path}')
                    elif gen == Generator.MBDYN:
                        codegen_mbdyn(model.output, infos,
                                      selected['main'], selected['trgt'], selected['trim'])
                        status.append(f'Saved {name} generator to {out_path}')
                if model.enabled_gens.count(True) == 0:
                    status.append('No output format enabled')
                messages.append((Message.STATUS, '\n'.join(status)))

        view(model)
        if len(messages) > 0:
            continue

        key = ord(getch())
        messages.append((Message.KEY, key))


MBD_PREFIX = 'LIDIA_'


def codegen_mbdyn(output: str, infos: Dict[str, FieldInfo],
                  main_fields: List[str], trgt_fields: List[str], trim_fields: List[str]) -> None:
    filename = os.path.basename(output)
    defines = ['output']
    for group, prefix in [(main_fields, ''), (trgt_fields, 'trgt_'), (trim_fields, 'trim_')]:
        for f in group:
            if issubclass(infos[f].field_type, BaseModel):
                defines.extend(
                    f'{prefix}{f}_{a}' for a in infos[f].field_type.__fields__)
            else:
                defines.append(f'{prefix}{f}')
    with open(output + '.set', 'w') as out_set:
        out_set.write('''##### Generated with pack_maker.py
# constants definition file
# include it in your model like this:
#
# include: "./{}.set";

### identifiers:
'''.format(filename))

        next_id = 55000
        for d in defines:
            out_set.write('set: const integer {}{} = {};\n'.format(
                MBD_PREFIX,
                d.upper(),
                next_id))
            next_id += 1
        out_set.write('''
# vim:ft=mbd
''')

    with open(output + '.drv', 'w') as out_drv:
        out_drv.write('''##### Generated with pack_maker.py
# drive definition file
# include it in your model like this:
#
# include: "./{}.drv";
#
# modify the following drive callers to get relevant data from model
# example of getting Z-axis position of NODE_CG:
#
# drive caller: LIDIA_NED_DOWN, node, NODE_CG, structural, string, "X[3]", direct;

### drive templates:
'''.format(filename))
        for d in defines[1:]:  # skip stream id
            out_drv.write('drive caller: {}{}, const, 0;\n'.format(
                MBD_PREFIX,
                d.upper()))
        out_drv.write('''
# vim:ft=mbd
''')

    with open(output + '.elm', 'w') as out_elm:
        out_elm.write('''##### Generated with pack_maker.py
# output stream definition
# include it in your model **inside the elements block** like this:
#
# include: "./{}.elm";
#
# you also need to increase the number of output elements in the control data block:
# begin: control data;
#     output elements: +1;
# end: control data;

### output stream definition
stream output: {}{},
\tstream name, "Lidia output in SMOL format",
\tcreate, no,
\tport, 5100,
\thost, "127.0.0.1",
\tsocket type, udp,
'''.format(filename, MBD_PREFIX, defines[0].upper()))

        values = StringIO()
        pack_fieldlists(Generator.MBDYN, values, infos,
                        main_fields, trgt_fields, trim_fields)

        out_elm.write('\tvalues, {},\n'.format(values.getvalue().count('\n')))
        out_elm.write(values.getvalue())
        out_elm.write('''
\tmodifier, copy cast,
\t\tswap, f, d, i,
''')
        type_out = StringIO()
        pack_fieldlists(Generator.MBDYN_TYPE, type_out, infos,
                        main_fields, trgt_fields, trim_fields)
        out_elm.write(type_out.getvalue()[:-2])  # strip ending ',\n'
        out_elm.write('''
\t;

# vim:ft=mbd
''')


def codegen_matlab(out: TextIOBase, function_name: str, infos: Dict[str, FieldInfo],
                   main_fields: List[str], trgt_fields: List[str], trim_fields: List[str]) -> None:
    out.write('function data = {}( ...\n    '.format(function_name))
    arglist = []
    for f in main_fields:
        if issubclass(infos[f].field_type, BaseModel):
            arglist.extend('{}_{}'.format(f, a)
                           for a in infos[f].field_type.__fields__)
        else:
            arglist.append(f)
    for f in trgt_fields:
        if issubclass(infos[f].field_type, BaseModel):
            arglist.extend('trgt_{}_{}'.format(f, a)
                           for a in infos[f].field_type.__fields__)
        else:
            arglist.append('trgt_{}'.format(f))
    for f in trim_fields:
        if issubclass(infos[f].field_type, BaseModel):
            arglist.extend('trim_{}_{}'.format(f, a)
                           for a in infos[f].field_type.__fields__)
        else:
            arglist.append('trim_{}'.format(f))

    out.write(', ...\n    '.join(arglist))
    out.write(''') % {} arguments
%PACK_LIDIA Pack aircraft data into binary format
%   The output is array of bytes in MsgPack format, as expected by
%   'smol' source of lidia package
%
%   This is generated using pack_maker.py to create code suitable for
%   use in Simulink - known size of I/O (see below), no map or struct usage
%
%   Arguments:
'''.format(len(arglist)))
    for fieldgroup, arg_prefix, doc_prefix in [(main_fields, '', ''), (trgt_fields, 'trgt_', 'TARGET '), (trim_fields, 'trim_', 'TRIM ')]:
        for f in fieldgroup:
            out.write('%       {}{}: {}{}\n'.format(
                arg_prefix, f, doc_prefix, infos[f].docstring))
    out.write('''
    data = uint8([...
''')
    length = pack_fieldlists(Generator.MATLAB, out,
                             infos, main_fields, trgt_fields, trim_fields)

    out.write('''    ]);
% data length {0} bytes
end

% You might need to manually configure Simulink output:
%   Size: 1 {0}
%   Type: uint8
% The warning message "Wrap on overflow detected" can be ignored

function bytes = b(value)
    % reverse byte order to convert from little endian to big endian
    bytes = typecast(swapbytes(value), 'uint8');
end
'''.format(length))


def pack_fieldlists(gen: Generator, out: TextIOBase, infos: Dict[str, FieldInfo], main_fields, trgt_fields, trim_fields) -> int:
    length = 0

    field_count = len(main_fields) + \
        (1 if len(trgt_fields) > 0 else 0) + \
        (1 if len(trim_fields) > 0 else 0)
    length += pack_map(gen, out, field_count)

    for function_name in main_fields:
        length += pack_field(gen, out, infos[function_name], '')

    for fieldgroup, prefix in [(trgt_fields, 'trgt'), (trim_fields, 'trim')]:
        if len(fieldgroup) > 0:
            length += pack_str(gen, out, prefix)
            length += pack_map(gen, out, len(fieldgroup))
            for function_name in fieldgroup:
                length += pack_field(gen, out, infos[function_name], prefix)
    return length


def pack_field(gen: Generator, out: TextIOBase, info: FieldInfo, prefix: str) -> int:
    length = 0
    length += pack_str(gen, out, info.name)
    if issubclass(info.field_type, VectorModel):
        length += pack_vector(gen, out, info.name, info.field_type, prefix)
    elif info.field_type is int:
        length += pack_int(gen, out, info.name if len(prefix) ==
                           0 else '{}_{}'.format(prefix, info.name))
    else:
        raise NotImplementedError(
            'No code generation for class:', info.field_type)
    return length


def pack_vector(gen: Generator, out: TextIOBase, name: str, cls: Type[VectorModel], prefix: str) -> int:
    length = 0
    length += pack_array(gen, out, len(cls.__fields__))
    for inner in cls.__fields__:
        length += pack_float(gen, out, '{}{}_{}'.format(
            '' if len(prefix) == 0 else (prefix + '_'),
            name, inner), double=(name == 'ned'))  # special case for positional accuracy
    return length


def pack_map(gen: Generator, out: TextIOBase, count: int) -> int:
    if count > 15:
        raise NotImplementedError('only fixmap handled for now')
    if gen == Generator.MATLAB:
        out.write(
            '        0x8{:x}, ... % map length {}\n'.format(count, count))
    elif gen == Generator.MBDYN:
        val = 0x80 + count
        out.write(
            f'\tdrive, const, {val},\t# 0x{val:02X} map length {count}\n')
    elif gen == Generator.MBDYN_TYPE:
        out.write('\t\tuint8_t,\n')
    return 1


def pack_array(gen: Generator, out: TextIOBase, count: int) -> int:
    if count > 15:
        raise NotImplementedError('only fixarray handled for now')
    if gen == Generator.MATLAB:
        out.write(
            '        0x9{:x}, ... % array length {}\n'.format(count, count))
    elif gen == Generator.MBDYN:
        val = 0x90 + count
        out.write(
            f'\tdrive, const, {val},\t# 0x{val:02X} array length {count}\n')
    elif gen == Generator.MBDYN_TYPE:
        out.write('\t\tuint8_t,\n')

    return 1


def pack_str(gen: Generator, out: TextIOBase, data: str) -> int:
    if len(data) > 31:
        raise NotImplementedError('only fixstr handled for now')
    if gen == Generator.MATLAB:
        out.write("        0x{:x}, '{}', ... % string length {}\n".format(
            0b10100000 + len(data), data, len(data)))
    elif gen == Generator.MBDYN:
        val = 0b10100000 + len(data)
        out.write(
            f'\tdrive, const, {val},\t# 0x{val:02X} string length {len(data)}\n')
        for c in data:
            out.write(f"\tdrive, const, {ord(c)},\t# '{c}'\n")
    elif gen == Generator.MBDYN_TYPE:
        for _ in range(len(data) + 1):
            out.write('\t\tuint8_t,\n')
    return 1 + len(data)


def pack_float(gen: Generator, out: TextIOBase, name: str, double=False) -> int:
    if gen == Generator.MATLAB:
        out.write('        0xc{}, b({}({})), ... % {}\n'.format(
            'b' if double else 'a',
            'double' if double else 'single',
            name,
            'double' if double else 'float',
        ))
    elif gen == Generator.MBDYN:
        val = 0xCB if double else 0xCA
        out.write('\tdrive, const, {},\t# 0x{:02X} {}\n'.format(
            val, val, 'double' if double else 'float'
        ))
        out.write(f'\tdrive, reference, {MBD_PREFIX}{name.upper()},\n')
    elif gen == Generator.MBDYN_TYPE:
        out.write('\t\tuint8_t,\n')
        out.write(f'\t\t{"double" if double else "float"},\n')
    return 9 if double else 5


def pack_int(gen: Generator, out: TextIOBase, name: str) -> int:
    if gen == Generator.MATLAB:
        out.write(
            '        0xce, b(uint32({})), ... % 32-bit unsigned integer\n'.format(name))
    elif gen == Generator.MBDYN:
        val = 0xCE
        out.write(f'\tdrive, const, {val},\t# 0x{val:02X} uint32_t\n')
        out.write(f'\tdrive, reference, {MBD_PREFIX}{name.upper()},\n')
    elif gen == Generator.MBDYN_TYPE:
        out.write('\t\tuint8_t,\n')
        out.write('\t\tuint32_t,\n')
    return 5


if __name__ == '__main__':
    main()



"""listen to UDP packets from FlightGear native socket protocol"""
from argparse import _SubParsersAction, ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from multiprocessing import Queue
import socket
from typing import Tuple

from ..aircraft import *
from ..config import Config
from .mytypes import RunFn, FGNetFDM


def setup(subparsers: _SubParsersAction) -> Tuple[str, RunFn]:
    NAME = 'flightgear'
    parser: ArgumentParser = subparsers.add_parser(
        NAME,
        help=__doc__,
        formatter_class=ArgumentDefaultsHelpFormatter,
        epilog="""To send packets at 30Hz from a local FlightGear instance,
add this command line argument: --native-fdm=socket,out,30,127.0.0.1,5004,udp""")
    parser.add_argument('-i', '--ip',
                        help='listen UDP adress', default='0.0.0.0')
    parser.add_argument('-p', '--port', type=int,
                        help='listen UDP port', default=5004)

    return (NAME, run)


def run(q: Queue, args: Namespace, config: Config):
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
        # Timeout is required to handle leaving with Ctrl+C
        sock.settimeout(1.0)
        sock.bind((args.ip, args.port))
        if args.verbosity >= 0:
            print(f'Listening for UDP packets on {args.ip}:{args.port}')

        while True:
            try:
                data, _ = sock.recvfrom(1024)
                fg = FGNetFDM.from_bytes(data)

                FT = 0.3048
                """conversion multiplier from feet to meters"""

                state = AircraftState()
                state.ned = NED()
                # TODO: Add argument for reference location, calculate north and east from that
                state.ned.down = -fg.altitude
                state.att = Attitude()
                state.att.roll = fg.phi
                state.att.pitch = fg.theta
                state.att.yaw = fg.psi
                # Velocities in ft/s
                state.v_body = XYZ()
                state.v_body.x = fg.v_body_u * FT
                state.v_body.y = fg.v_body_v * FT
                state.v_body.z = fg.v_body_w * FT
                state.v_ned = NED()
                state.v_ned.north = fg.v_north * FT
                state.v_ned.east = fg.v_east * FT
                state.v_ned.down = fg.v_down * FT
                state.ctrl = Controls()
                state.ctrl.stick_pull = -fg.elevator
                state.ctrl.stick_right = fg.right_aileron
                state.ctrl.pedals_right = fg.rudder

                state.model_instruments(config)
                if state.instr.ralt is not None:
                    state.instr.ralt = fg.agl * config.instruments.altitude_multiplier
                state.set_time(config)

                q.put(('smol', state.smol()))
            except socket.timeout:
                continue
